VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "ACEFile"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Description = "MSTS Shape File Model Object"
Option Explicit
Dim comp As CompressZIt

Public dx As D3DX8
Attribute dx.VB_VarMemberFlags = "40"
Public textureSearchPath As String
Attribute textureSearchPath.VB_VarDescription = "If a texture is not found in the same folder as the S file, this ; delimited string is scanned."


Public Function loadAce(ByVal fName As String, d3dd As Direct3DDevice8) As Direct3DTexture8
Attribute loadAce.VB_Description = "Load ACE file into Texture8 structure"
    Dim b() As Byte
    Dim hdr(6) As Long, siz As Long
    Dim tx As Direct3DTexture8
    Dim top As Direct3DSurface8
    Dim data() As Byte
    Dim bufLen As Long
    Dim bOffset As Long
    Dim srect As RECT
    Dim fmt As CONST_D3DFORMAT
    Dim sfmt As CONST_D3DFORMAT
    Dim i As Long, iRGB As Long, j As Long, iPix As Long
    Dim pitch As Long
    
    If fName = "" Then Exit Function
    
    ' file not there ?
    If Dir(fName) = "" Then
        'CHECK SEARCH PATH
        If textureSearchPath = "" Then Exit Function
        Dim s() As String
        s = Split(textureSearchPath, ";")
        For i = 0 To UBound(s)
            If Dir(s(i) & fName) <> "" Then Exit For
        Next
        If i > UBound(s) Then Exit Function
        fName = s(i) & fName
    End If
    
    If readFile(fName, b) = 0 Then
    
        DXCopyMemory hdr(0), b(0), 28
        
        ' mip table offset
        bOffset = 152 + hdr(5) * 16
        
        ' ignore cab ACE files
        If hdr(1) = 0 And hdr(3) <> hdr(2) Then Exit Function
        
        If hdr(1) > 15 Then
            fmt = D3DFMT_DXT1
            sfmt = findTexFormat(fmt, d3dd)
            
            pitch = hdr(2) * 2                  ' DXT1 is a 16 bit format
            DXCopyMemory bufLen, b(bOffset), 4
            DXCopyMemory siz, b(bufLen), 4
            ReDim data(siz)
            DXCopyMemory data(0), b(bufLen + 4), siz
        
        Else
            siz = hdr(2) * hdr(3)
            ReDim data(siz * 4)
            fmt = D3DFMT_A8R8G8B8
            sfmt = findTexFormat(fmt, d3dd)
            
            pitch = hdr(2) * 4
            
            ' i=0 to height-1
            For i = 0 To hdr(3) - 1
            
                iPix = i * hdr(2) * 4
                ' 3 sets of RGB
                DXCopyMemory bufLen, b(bOffset), 4
                
                For iRGB = 0 To 2
                    ' single row of RGB to width
                    For j = 0 To hdr(2) - 1
                        data(iPix + j * 4 + 2 - iRGB) = b(bufLen + j + iRGB * hdr(2))
                    Next
                Next
                bOffset = bOffset + 4
                
                ' alpha - there appears to be an extra width\8 bytes (transparency?) before this
                If hdr(5) = 5 Then
                    bufLen = bufLen + 3 * hdr(2) + hdr(2) \ 8
                    ' single row of alpha to width
                    For j = 0 To hdr(2) - 1
                        data(iPix + j * 4 + 3) = b(bufLen + j)
                    Next
                    
                ' transparency
                ElseIf hdr(5) = 4 Then
                    bufLen = bufLen + 3 * hdr(2)
                    iRGB = 1
                    ' single row of RGB to width
                    For j = 0 To hdr(2) - 1 Step 8
                        ' small optimise
                        If b(bufLen) <> 0 Then
                            ' scan bits...
                            For iRGB = 7 To 0 Step -1
                                If (b(bufLen) Mod 2) = 1 Then data(iPix + (j + iRGB) * 4 + 3) = 255
                                b(bufLen) = b(bufLen) \ 2
                            Next
                        End If
                        bufLen = bufLen + 1
                    Next
                    
                Else  ' opaque
                    iRGB = iPix + hdr(2) * 4 - 1
                    For j = iPix + 3 To iRGB Step 4
                        data(j) = 255
                    Next
                End If
                
            Next
        End If
        
        Set tx = d3dd.CreateTexture(hdr(2), hdr(3), 6, 0, sfmt, D3DPOOL_MANAGED)
        
        srect.bottom = hdr(3)
        srect.Right = hdr(2)
        
        ' if card does not support dxt, this should do the conversion
        For j = 0 To 5
            Set top = tx.GetSurfaceLevel(j)
            dx.LoadSurfaceFromMemory top, ByVal 0, ByVal 0, data(0), fmt, pitch, ByVal 0, srect, D3DX_FILTER_LINEAR, 0
        Next
        Set loadAce = tx
    Else
        Set loadAce = Nothing
    End If

End Function

Private Function findTexFormat(ByVal fmt As CONST_D3DFORMAT, d3dd As Direct3DDevice8) As CONST_D3DFORMAT
    Dim d3d As Direct3D8
    Dim dMode As D3DDISPLAYMODE
    
    Set d3d = d3dd.GetDirect3D
    d3d.GetAdapterDisplayMode D3DADAPTER_DEFAULT, dMode
    
    ' DXT?
    If fmt <> D3DFMT_A8R8G8B8 And d3d.CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, dMode.Format, 0, D3DRTYPE_TEXTURE, fmt) = D3D_OK Then
        findTexFormat = fmt
        Exit Function
    End If
    
    ' 32 BIT
    fmt = D3DFMT_A8R8G8B8
    If d3d.CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, dMode.Format, 0, D3DRTYPE_TEXTURE, fmt) = D3D_OK Then
        findTexFormat = fmt
        Exit Function
    End If
    
    ' 16 BIT
    findTexFormat = D3DFMT_A4R4G4B4
    
End Function
' read MSTS file into byte array
' works if GZip compressed too
Public Function readFile(fName As String, ByRef bData() As Byte) As Long
Attribute readFile.VB_Description = "Returns uncompressed file in a byte array. Return values >0 indicate an error."
    Dim i As Integer
    Dim bufSize As Long
    Dim bHead(7) As Byte
    Dim res As Long
    
    i = FreeFile
    On Error GoTo ErrEx
    
    Open fName For Binary Access Read As i
    Get #i, , bHead()
    ReDim bData(LOF(i) - 17)
    Get #i, 17, bData()
    If bHead(7) > 64 Then
        Get #i, , bufSize
        Set comp = New CompressZIt
        readFile = comp.DecompressData(bData(), bufSize)
        Set comp = Nothing
    End If
    Close i
    Exit Function
ErrEx:
    readFile = Err.Number
End Function

Private Sub Class_Initialize()
    Set dx = New D3DX8
End Sub

Private Sub Class_Terminate()
    Set dx = Nothing
End Sub
